SystemOrganization addCategory: #'CommandShell-Piping'!

OrderedCollection subclass: #ProxyPipeline
	instanceVariableNames: 'commandLine background completionSemaphore completionWatcher'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Piping'!

!ProxyPipeline commentStamp: 'dtl 12/13/2002 19:30' prior: 0!
A collection of connected PipeJunctions, representing external OS processes or internal evaluators. This class exists primarily to make the functioning of a collection of command pipelines, some of which may be evaluated as asynchronous "background" processes, easier to understand.

Events triggered by my proxies are handled and forwarded in such a way that a client (such as a CommandShell) will receive events from a ProxyPipeline as if it were an individual PipeJunction.

The user of a ProxyPipeline is responsible for closing the external resources associated with the proxies by sending either #closePipes or #finalize.!

!ProxyPipeline class methodsFor: 'command processing' stamp: 'dtl 4/27/2003 11:32'!
command: aCommandString
	"Evaluate a new instance created from aCommandString. Sender is responsible
	for closing the pipes with #closePipes or #finalize."

	"ProxyPipeline command: 'ls | cat | wc'"
	"ProxyPipeline command: 'ls NOSUCHFILE * | cat | wc'"
	"ProxyPipeline command: 'ls | copyToOutput | wc'"
	"ProxyPipeline command: ''"

	^ self command: aCommandString shell: CommandShell new
! !

!ProxyPipeline class methodsFor: 'command processing' stamp: 'dtl 4/27/2003 11:32'!
command: aCommandString shell: aCommandShell
	"Evaluate a new instance created from aCommandString using aCommandShell.
	Sender is responsible for closing the pipes #closePipes or #finalize."

	"ProxyPipeline command: 'ls | cat | wc' shell: CommandShell new"
	"ProxyPipeline command: 'ls NOSUCHFILE * | cat | wc' shell: CommandShell new"
	"ProxyPipeline command: 'ls | copyToOutput | wc' shell: CommandShell new "
	"ProxyPipeline command: '' shell: CommandShell new"

	^ (self fromString: aCommandString shell: aCommandShell) value
! !

!ProxyPipeline class methodsFor: 'instance creation' stamp: 'dtl 6/8/2006 06:57'!
fromString: aCommandString shell: aCommandShell
	"Answer a new instance created from aCommandString using aCommandShell."

	"ProxyPipeline fromString: 'ls | cat | wc' shell: CommandShell new"
	"ProxyPipeline fromString: 'ls NOSUCHFILE * | cat | wc' shell: CommandShell new"
	"ProxyPipeline fromString: 'ls | copyToOutput | wc' shell: CommandShell new "
	"ProxyPipeline fromString: '' shell: CommandShell new"

	^ super new fromString: aCommandString shell: aCommandShell
! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 3/29/2003 13:49'!
atEnd

	^ self last atEnd
! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:42'!
background
	"Answer true if this pipeline should be evaluated in the background."

	^ background ifNil: [background := false]! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:42'!
background: trueOrFalse

	background := trueOrFalse! !

!ProxyPipeline methodsFor: 'initialize - release' stamp: 'dtl 12/31/2002 10:57'!
close
	"Close input pipe to first proxy in the pipeline. The proxies are not
	finalized, this simply closes the input stream to allow normal processing
	to proceed to completion."

	self isEmpty ifFalse: [self first close]
! !

!ProxyPipeline methodsFor: 'initialize - release' stamp: 'dtl 9/3/2010 17:37'!
closePipes

	| p |
	self do: [:e | e closePipes].
	self unregisterEvents.
	(p := self pipeFromOutput) ifNotNil: [p close]
! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/20/2002 19:20'!
commandLine
	"Command line string from which this pipeline was constructed"

	^ commandLine! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:42'!
commandLine: aString
	"Command line string from which this pipeline was constructed"

	commandLine := aString! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/20/2006 13:28'!
completionSemaphore
	"Answer the value of completionSemaphore. This semaphore is signaled
	when the last proxy completes. Subsequent cleanup is preformed by a
	process waiting on the semaphore. Completion notification is often triggered
	by the grimReaperProcess, so the separate pipeline cleanup prevents
	those activities from being processed in the context of the grimReaperProcess."

	^ completionSemaphore ifNil: [completionSemaphore := Semaphore new]! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/20/2006 13:21'!
completionWatcher
	"A process that waits for notification that the last proxy has completed,
	and performs any necessary cleanup. Completion notification is often
	triggered by the grimReaperProcess, so the separate pipeline cleanup
	prevents those activities from being processed in the context of the
	grimReaperProcess."

	^ completionWatcher! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:42'!
completionWatcher: anObject
	"Set the value of completionWatcher"

	completionWatcher := anObject! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 4/28/2003 21:02'!
errorPipelineStream
	"Accumulates the error output of commands in a command pipeline."

	self isEmpty
		ifTrue: [^ nil]
		ifFalse: [^ self last errorPipelineStream]! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:10'!
errorUpToEnd
	"Answer all available characters from the error stream shared by my proxies."

	^ self last errorUpToEnd
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 8/25/2006 12:22'!
errorUpToEndOfFile
	"Answer all available characters from the error stream. Block and
	continue reading until end of file is detected."

	^ self last errorUpToEndOfFile
! !

!ProxyPipeline methodsFor: 'finalization' stamp: 'dtl 4/9/2006 09:43'!
finalize

	self closePipes.
	^ super finalize
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:10'!
flush
	"Flush output to the standard input stream of my first proxy."

	^ self first flush! !

!ProxyPipeline methodsFor: 'initialize - release' stamp: 'dtl 11/8/2007 20:42'!
fromString: aCommandString shell: aCommandShell 
	"Initialize a new instance created from aCommandString using aCommandShell. "

	self commandLine: aCommandString.
	(aCommandShell splitPipelineCommands: aCommandString)
		inject: nil
		into: [:prevProxy :command | 
			| nextProxy |
			nextProxy := aCommandShell
				redirectedPipeableProxyFor: command
				predecessorProxy: prevProxy.
			prevProxy
				ifNotNil: [prevProxy prepareOutputFor: nextProxy.
					prevProxy canProvideOutputPipe
						ifFalse: [nextProxy closeWriter]].
			self add: nextProxy].
	self isEmpty
		ifFalse: [self last prepareOutputFor: self; addDependent: self]! !

!ProxyPipeline methodsFor: 'event handling' stamp: 'dtl 9/3/2010 17:51'!
handleCompletionWhenSignaled: aSemaphore
	"Answer a process that will complete processing the the last proxy has
	signaled its completion."

	| lastProxy proc |
	proc := [aSemaphore wait.
		self waitForAllToComplete.
		(lastProxy := self last)
			ifNotNil: [lastProxy closeErrorPipeline.
					self triggerEvent: #complete]] newProcess.
	^ proc resume! !

!ProxyPipeline methodsFor: 'event handling' stamp: 'dtl 9/3/2010 17:40'!
handleRunstateChange

	| lastProxy |
	(lastProxy := self last)
		ifNotNil: [lastProxy isComplete
					ifTrue: [self completionSemaphore signal]]! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 9/12/2003 11:53'!
isComplete
	"True if all proxies are complete. The proxies may report their completion events
	out of sequence, so test all the proxies (not just the last one in the pipeline)."

	^ self noneSatisfy: [:proxy | proxy isComplete not]
! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 4/19/2003 09:16'!
isExternalProcess
	"Answer true if the process which I represent is an external OSProcess. For protocol
	compatibility with PipeJunction."

	^ false! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 11/23/2002 13:37'!
needsPrompt

	^ self last needsPrompt! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:08'!
next

	^ self last next
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:08'!
next: count

	^ self last next: count
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:11'!
nextFromError: count
	"Answer up to count characters from the error pipeline stream, or an empty string
	if no data is available. All characters are routed through the errorPipelineStream,
	which is shared by all my proxies."

	^ self last next: count
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:09'!
nextPut: aCharacter
	"Write aCharacter to the standard input stream of my first proxy."

	^ self first nextPut: aCharacter! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 11/21/2002 18:09'!
nextPutAll: characters
	"Write characters to the standard input stream of my first proxy."

	^ self first nextPutAll: characters! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 4/14/2006 10:32'!
output

	^ self last output! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 4/29/2003 05:04'!
pipeFromOutput
	"The output from the last proxy in the pipeline, if any"

	self isEmpty
		ifTrue: [^ nil]
		ifFalse: [^ self last pipeFromOutput]! !

!ProxyPipeline methodsFor: 'accessing' stamp: 'dtl 4/28/2003 20:58'!
pipeToInput
	"The input to the first proxy in the pipeline, if any"

	self isEmpty
		ifTrue: [^ nil]
		ifFalse: [^ self first pipeToInput]! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 9/12/2003 11:49'!
pipesAreEmpty
	"True after processing is complete and all available data has been read from
	the output pipe and the errorPipelineStream pipe."

	^ self isComplete and: [self last pipesAreEmpty]
! !

!ProxyPipeline methodsFor: 'printing' stamp: 'dtl 11/20/2002 19:32'!
printOn: aStream

	self background
		ifTrue: [aStream nextPutAll: 'a background ']
		ifFalse: [aStream nextPutAll: 'a foreground '].
	aStream nextPutAll: self class name, ' for "', self commandLine asString, '"'! !

!ProxyPipeline methodsFor: 'initialize - release' stamp: 'dtl 1/4/2003 15:35'!
release

	self do: [:proxy | proxy release].
	^ super release! !

!ProxyPipeline methodsFor: 'testing' stamp: 'dtl 7/5/2006 09:07'!
succeeded
	"Answer true if all proxies succeeded, otherwise false"

	self detect: [:proxy | proxy succeeded not] ifNone: [^ true].
	^ false
! !

!ProxyPipeline methodsFor: 'initialize - release' stamp: 'dtl 8/20/2006 19:05'!
unregisterEvents

	self isEmpty ifFalse: [self last removeDependent: self]
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 8/25/2006 12:21'!
upToEnd
	"Answer all available characters from the output stream."

	^ self last upToEnd
! !

!ProxyPipeline methodsFor: 'streaming' stamp: 'dtl 8/25/2006 12:21'!
upToEndOfFile
	"Answer all available characters from the output stream. Block and
	continue reading until end of file is detected."

	^ self last upToEndOfFile
! !

!ProxyPipeline methodsFor: 'updating' stamp: 'dtl 8/22/2006 06:28'!
update: aParameter

	aParameter == self pipeFromOutput
		ifTrue: [^ self triggerEvent: #outputDataReady].
	aParameter == self errorPipelineStream
		ifTrue: [^ self triggerEvent: #errorDataReady].
	aParameter == #runState
		ifTrue: [^ self handleRunstateChange].
	self error: 'unexpected parameter'

! !

!ProxyPipeline methodsFor: 'evaluation' stamp: 'dtl 11/25/2006 11:01'!
value
	"Initiate evaluation of each member of the pipeline, and answer the
	last proxy in the pipeline. Evaluation may proceed asynchronously, and
	the sender should wait for the last proxy to complete its evalation in order
	to obtain complete output and error contents from the pipeline."

	"(ProxyPipeline command: 'ls | cat | wc' shell: CommandShell new) value"

	self completionWatcher: (self handleCompletionWhenSignaled: self completionSemaphore).
	self do: [:proxy | proxy value]
! !

!ProxyPipeline methodsFor: 'event handling' stamp: 'dtl 11/20/2006 12:43'!
waitForAllToComplete
	"In some cases the last proxy in a pipeline may complete before some
	of the others. In particular, if one proxy has redirected its output to
	a file, the next proxy will see a nil input stream, and may quickly
	complete its processing before its predecessor proxies have finished
	writing to the file. Time out with an error if pipeline fails to complete
	after 10 seconds."

	(1 to: 100) do: [:e |
		self isComplete ifTrue: [^ self].
		(Delay forMilliseconds: 100) wait].
	self error: 'pipeline did not complete evaluation'

! !

Object subclass: #RemoteTask
	instanceVariableNames: 'task childProxy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Piping'!

!RemoteTask commentStamp: 'dtl 1/25/2010 20:19' prior: 0!
RemoteTask do: [2 + 2]

A RemoteTask represents a block to be evaluated in a headless Squeak child process. The results of evaluating the block are returned to the sender through a reference stream on an OS pipe.

The block is evaluated in a remote headless image beginning with a clone of the sender image at the time of the message send. All side effects of evaluating the task block are localized to the remote image, and have no effect on the sending image. The result object may be an object of any complexity, such as a dictionary or array of result values.

On a unix system, the creation of child Squeak images is a relatively light weight operation, so tasks of varying degrees of complexity can be assigned to remote tasks with only moderate overhead.

!

!RemoteTask class methodsFor: 'evaluating' stamp: 'dtl 1/25/2010 20:51'!
do: aBlock
	"Start a remote task and read the result object."

	"RemoteTask do: [2 + 2]"

	^(self task: aBlock) value! !

!RemoteTask class methodsFor: 'evaluating' stamp: 'dtl 1/25/2010 20:50'!
start: aBlock
	"Create a remote task and start evaluation. Do not read the result object."

	^(self task: aBlock) start! !

!RemoteTask class methodsFor: 'instance creation' stamp: 'dtl 1/25/2010 19:03'!
task: aBlock

	^ self new task: aBlock! !

!RemoteTask methodsFor: 'evaluating' stamp: 'dtl 8/16/2010 07:18'!
result
	"Answer the result of a previously started task. The child Squeak sends one
	result object back to its parent Squeak before exiting."

	| s responseObject responseStream pollingDelay |
	"Wait for child to complete. Prevents this image from looping on empty stream."
	pollingDelay := Delay forMilliseconds: 50.
	[childProxy isComplete] whileFalse: [pollingDelay wait].
	"Read the result object from the child Squeak process."
	s := RWBinaryOrTextStream with: childProxy pipeFromOutput upToEndOfFile.
	s reset.
	responseStream := ReferenceStream on: s.
	responseObject := responseStream next.
	s close.
	^ responseObject
! !

!RemoteTask methodsFor: 'evaluating' stamp: 'dtl 1/25/2010 21:09'!
start
	"Fork a headless child Squeak to evaluate my task."

	| childBlock out result |
	"Create block to be evaluated in a Squeak child process."
	childBlock := ["First set standard input to blocking. This avoids a delay to wait for
		data to be available, but note that it is dangerous to do this unless you
		are certain that you know in advance that the correct number of characters
		will be available on the input. If this is not the case, then the VM will block
		on the read and lock up the child squeak process."
		OSProcess thisOSProcess stdIn setBlocking.
		out := ReferenceStream on: OSProcess thisOSProcess stdOut.
		result := [self task value]
					on: Exception
					do: [:e | result := 'remote task execution failed: ', e printString].
		out nextPut: result].
	"Fork a child Squeak that will evaluate the childBlock then exit"
	childProxy := PipeableOSProcess forkHeadlessSqueakAndDoThenQuit: childBlock
! !

!RemoteTask methodsFor: 'accessing' stamp: 'dtl 1/25/2010 19:03'!
task
	"Answer the value of task"

	^ task! !

!RemoteTask methodsFor: 'accessing' stamp: 'dtl 1/25/2010 19:03'!
task: anObject
	"Set the value of task"

	task := anObject! !

!RemoteTask methodsFor: 'evaluating' stamp: 'dtl 1/25/2010 20:50'!
value
	"Start a remote task and read the result object. The child Squeak sends one
	result object back to its parent Squeak before exiting."

	"(RemoteTask task: [2 + 2]) value"

	self start.
	^self result
! !

PipeJunction subclass: #PipeableEvaluator
	instanceVariableNames: 'argumentArray runState exitStatus outputHasBeenRedirected'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Piping'!

!PipeableEvaluator commentStamp: 'dtl 2/1/2003 16:56' prior: 0!
I evaluate Smalltalk expressions, taking input from my pipeToInput, and print the results of the evaluation on my pipeFromOutput. I may append error messages to my errorPipelineStream. I provide a framework for pipelines of expressions, possibly combined in the same pipeline with ExternalOSProcess proxies executing external operating system commands.

My evaluationBlock may be supplied directly, or can be compiled from an expression string. If compiled from an expression string, the names 'stdin', 'stdout', 'stderr', and 'args' are used to refer to the input stream, output stream, error stream, and an array of arguments. If any of these names are used in the expression string, then the evaluation block will be compiled in such as way as to connect these names with their corresponding streams or argument array.
!

PipeableEvaluator subclass: #ExpressionEvaluator
	instanceVariableNames: 'expression evaluationBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Piping'!

!ExpressionEvaluator commentStamp: 'dtl 2/1/2003 16:58' prior: 0!
I provide a context in which a Smalltalk expression can be evaluated as an element of a command pipeline.!

!ExpressionEvaluator class methodsFor: 'instance creation' stamp: 'dtl 3/26/2006 12:10'!
block: aBlock

	"(self block: [#'This block answers a symbol as String']) value upToEnd"

	^ self basicNew block: aBlock
! !

!ExpressionEvaluator class methodsFor: 'instance creation' stamp: 'dtl 3/26/2006 12:09'!
command: aString

	"(self command: '''Hello world!!''') value upToEnd"

	^ self basicNew command: aString
! !

!ExpressionEvaluator class methodsFor: 'instance creation' stamp: 'dtl 12/13/2007 21:38'!
command: aString pipeFrom: inputStream errorPipelineStream: errorStream

	"(self command: 'stdout nextPutAll: stdin upToEnd'
		pipeFrom: (ReadStream on: 'Here is some input text')
		errorPipelineStream: InternalPipe nonBlockingPipe)
			value upToEnd"

	^ self command: aString pipeFrom: inputStream pipeTo: nil errorPipelineStream: errorStream! !

!ExpressionEvaluator class methodsFor: 'instance creation' stamp: 'dtl 12/13/2007 21:38'!
command: aString pipeFrom: inputStream pipeTo: outputStream errorPipelineStream: errorStream

	"(self command: 'stdout nextPutAll: stdin contents'
		pipeFrom: (ReadStream on: 'Here is some input text')
		pipeTo: nil
		errorPipelineStream: InternalPipe nonBlockingPipe)
			value upToEnd"

	^ super new
		expression: aString;
		pipeToInput: inputStream;
		pipeFromOutput: outputStream;
		errorPipelineStream: errorStream;
		notYetRunning;
		initialize! !

!ExpressionEvaluator methodsFor: 'evaluating' stamp: 'dtl 2/1/2003 16:34'!
addTreeLeaves: treeOfSymbols to: aCollection
	"Flatten a tree of Symbols built by a Scanner."

	treeOfSymbols do: [:e |
		(e isKindOf: Array)
			ifTrue: [self addTreeLeaves: e to: aCollection]
			ifFalse: [aCollection add: e]].
	^ aCollection! !

!ExpressionEvaluator methodsFor: 'initialize - release' stamp: 'dtl 4/19/2003 16:29'!
block: aBlock

	self evaluationBlock: aBlock.
	self initialize.
	self prepareOutputForInternalReader
! !

!ExpressionEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/8/2007 20:32'!
blockString
	"Answer source code for a block to execute my expression. Attempt to give it the correct
	number of parameters based on scanning expression for use of #args, #stderr, #stdout,
	and #stdin tokens."

	"PipeableEvaluator new expression: 'stdout nextPutAll: (args at: 4)'; blockString"
	"PipeableEvaluator new expression: 'stderr nextPutAll: stdin contents'; blockString"
	"PipeableEvaluator new expression: 'stdout nextPutAll: stdin contents'; blockString"
	"PipeableEvaluator new expression: 'stdin contents'; blockString"
	"PipeableEvaluator new expression: '#HelloWorld'; blockString"

	| tokens prefix |
	self expression isNil
		ifTrue:
			[^ '[]']
		ifFalse:
			[tokens := self addTreeLeaves: (Scanner new scanTokens: expression) to: OrderedCollection new.
			(tokens includes: #args)
				ifTrue:
					[prefix := '[:stdin :stdout :stderr :args | ']
				ifFalse:
					[(tokens includes: #stderr)
						ifTrue:
							[prefix := '[:stdin :stdout :stderr | ']
						ifFalse:
							[(tokens includes: #stdout)
								ifTrue:
									[prefix := '[:stdin :stdout | ']
								ifFalse:
									[(tokens includes: #stdin)
										ifTrue:
											[prefix := '[:stdin | ']
										ifFalse:
											[prefix := '[']]]].
			^ prefix, self expression, ']']
! !

!ExpressionEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/8/2007 20:32'!
blockValue
	"Evaluate my evaluationBlock. For a smart block, answer self. For a dumb
	block, answer the result of evaluating it. Heuristic: If the block receives
	two or more arguments, then assume that it was smart enough to write
	output to stdout. Otherwise, assume that it was a dumb block."

	| nArgs |
	self evaluationBlock ifNil: [^ nil].
	nArgs := evaluationBlock numArgs.
	(nArgs == 0)
		ifTrue: "dumb block"
			[^ evaluationBlock value]
		ifFalse:
			[(nArgs == 4)
				ifTrue: "smart block"
					[evaluationBlock
						value: self pipeToInput
						value: self pipeFromOutput
						value: self errorPipelineStream
						value: self argumentArray]
				ifFalse:
					[(nArgs == 3)
						ifTrue: "smart block"
							[evaluationBlock
								value: self pipeToInput
								value: self pipeFromOutput
								value: self errorPipelineStream]
						ifFalse:
							[(nArgs == 2)
								ifTrue: "smart block"
									[evaluationBlock
										value: self pipeToInput
										value: self pipeFromOutput]	
								ifFalse:
									[(nArgs == 1)
										ifTrue: "dumb block"
											[^ evaluationBlock value: self pipeToInput]
										ifFalse: "really stupid block"
											[self error: 'expected block with 4 or fewer arguments'.
											^ nil]]]]]
! !

!ExpressionEvaluator methodsFor: 'initialize - release' stamp: 'dtl 4/19/2003 16:25'!
command: aString

	self expression: aString.
	self initialize.
	self prepareOutputForInternalReader
! !

!ExpressionEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/29/2010 23:33'!
compile
	"Compile my expression into a block with zero to four parameters. The parameters,
	if present, correspond to my pipeToInput, pipeFromOutput, errorPipelineStream, and
	an array of extra parameters (in that order). By convention, these are named 'stdin',
	'stdout', 'stderr', and 'args' within my expression string.

	If the compilation fails due to a syntax error (compile time failure, not run time
	failure), put an error message on the standard error stream and answer a block
	which evaluates to a null string."

	evaluationBlock ifNil:
		[evaluationBlock :=
			[Compiler evaluate: self blockString
					for: self
					notifying: self
					logged: false]
				on: Exception
				do: [:e |
					self errorPipelineStream
						nextPutAll: 'Compiler: invalid Smalltalk expression ', self blockString;
						nextPut: Character lf.
					self fail.
					nil]]
! !

!ExpressionEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 16:34'!
evaluationBlock
	"Answer the block to be evaluated, possibly compiling it from the expression string."

	evaluationBlock ifNil: [self compile].
	^ evaluationBlock! !

!ExpressionEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:32'!
evaluationBlock: aBlock
	"The block to be evaluated, possibly compiled from the expression string."

	evaluationBlock := aBlock! !

!ExpressionEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 16:34'!
expression

	^ expression! !

!ExpressionEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:32'!
expression: aString

	expression := aString! !

!ExpressionEvaluator methodsFor: 'printing' stamp: 'dtl 3/24/2003 07:23'!
printOn: aStream

	super printOn: aStream.
	self expression isNil
		ifTrue:
			[self evaluationBlock isNil
				ifTrue:
					[aStream nextPutAll: ' for '.
					self evaluationBlock printOn: aStream]
				ifFalse:
					[aStream nextPutAll: ' for a '.
					self evaluationBlock class printOn: aStream]]
		ifFalse:
			[aStream nextPutAll: ' for '.
			self expression printOn: aStream]
! !

!ExpressionEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/8/2007 20:32'!
value
	"Evaluate my evaluationBlock. If the block evaluates to anything other
	than self, then print the result on pipeFromOutput as output. Otherwise,
	assume that the block was smart enough to have written its own output.
	For a block which does not write its own output, add a trailing line
	terminator after the evaluation if needed. Answer self as the result of
	the evaluation."

	| result resultString |
	self success. "Hope for the best, send #fail if something goes wrong"
	self isNotYetRunning ifTrue:
		[self ensurePipeFromOutput.
		self running.
		self evaluationBlock notNil
			ifTrue:
				[result := [self blockValue]
					on: Error
					do: [:e |
						self errorPipelineStream nextPutAll: e description; nextPut: Character lf.
						self fail].
				(result == self)
					ifFalse:
						[resultString := result asString.
						pipeFromOutput nextPutAll: resultString.
						(resultString isEmpty not and:
							[(resultString last ~= Character cr) and:
								[resultString last ~= Character lf]])
							ifTrue: [pipeFromOutput nextPut: Character lf]]]
			ifFalse:
				[self fail].
		self complete]
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 16:45'!
argumentArray

	^ argumentArray
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:36'!
argumentArray: anArrayOfStrings

	argumentArray := anArrayOfStrings
! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 6/8/2003 22:37'!
atEnd
	"True after processing is complete and all available data has been read from
	the output pipe."

	^ self isComplete
		and: [self outputHasBeenRedirected
				or: [self outputIsExternalStream
					or: [pipeFromOutput atEnd]]]
! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 11/15/2002 09:10'!
canProvideOutputPipe
	"Answer false if output has been redirected to a file and cannot be accessed by
	the next process proxy in a command pipeline."

	^ super canProvideOutputPipe and: [self outputHasBeenRedirected not]
! !

!PipeableEvaluator methodsFor: 'setting run state' stamp: 'dtl 11/24/2006 13:01'!
complete
	"Evaluation has been completed. Perform some cleanup and notify
	interested parties."

	self closePipeInputs.
	self runState: #complete.
	self errorHasBeenRedirected
		ifTrue: [self closeErrorPipeline]! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:36'!
errorPipelineStream
	"Accumulates the error output of commands in a command pipeline."

	errorPipelineStream isNil
		ifTrue:
			[errorPipelineStream := InternalPipe nonBlockingPipe.
			errorPipelineStream addDependent: self].
	^ errorPipelineStream
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:36'!
errorPipelineStream: aStream
	"Accumulates the error output of commands in a command pipeline."

	errorPipelineStream ifNotNil: [errorPipelineStream removeDependent: self].
	errorPipelineStream := aStream.
	errorPipelineStream ifNotNil: [errorPipelineStream addDependent: self]
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 4/19/2003 09:56'!
exitStatus

	^ exitStatus! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:36'!
exitStatus: aSymbol

	exitStatus := aSymbol! !

!PipeableEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/25/2001 22:39'!
fail

	self exitStatus: #fail! !

!PipeableEvaluator methodsFor: 'initialize - release' stamp: 'dtl 9/9/2002 13:51'!
initialize
	"If pipeFromOutput has been set, then the output stream is being redirected.
	Set a flag to indicate this, otherwise open new a pipe stream to use for output."

	pipeFromOutput isNil
		ifTrue: [self ensurePipeFromOutput]
		ifFalse: [self outputHasBeenRedirected: true].
	^ super initialize
! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 3/4/2002 06:58'!
isComplete

	^ self runState == #complete! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 3/4/2002 06:59'!
isNotYetRunning

	^ self runState == #notYetRunning! !

!PipeableEvaluator methodsFor: 'streaming' stamp: 'dtl 11/8/2007 20:37'!
nextPut: aCharacter
	"Write aCharacter to the standard input stream of my process."

	| readPosition |
	readPosition := pipeToInput position.
	pipeToInput setToEnd nextPut: aCharacter.
	pipeToInput position: readPosition.
	^ pipeToInput
! !

!PipeableEvaluator methodsFor: 'streaming' stamp: 'dtl 11/8/2007 20:37'!
nextPutAll: aCollection
	"Write aCollection to the standard input stream of my process."

	| readPosition |
	readPosition := pipeToInput position.
	pipeToInput setToEnd nextPutAll: aCollection.
	pipeToInput position: readPosition.
	^ pipeToInput
! !

!PipeableEvaluator methodsFor: 'setting run state' stamp: 'dtl 3/4/2002 22:32'!
notYetRunning
	"Process has not yet entered running state."

	self runState: #notYetRunning
! !

!PipeableEvaluator methodsFor: 'error handling' stamp: 'dtl 11/8/2007 20:37'!
notify: aString at: aPosition in: aStream

	| s |
	s := aStream contents.
	self errorPipelineStream
		nextPutAll: (s copyFrom: 1 to: (aPosition - 1));
		nextPutAll: aString;
		nextPutAll: (s copyFrom: aPosition to: s size);
		cr

! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:37'!
outputHasBeenRedirected
	"True if pipeFromOutput has been redirected to a FileStream"

	^ outputHasBeenRedirected ifNil: [outputHasBeenRedirected := false]! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:37'!
outputHasBeenRedirected: trueOrFalse
	"True if pipeFromOutput has been redirected to a FileStream"

	outputHasBeenRedirected := trueOrFalse! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 6/8/2003 22:35'!
outputIsExternalStream
	"True if output is being piped to an external proxy, or if output has been
	redirected to a file."

	^ self pipeFromOutput isKindOf: FileStream
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/29/2006 06:29'!
pipeFromError
	"Internal evaluators use the errorPipelineStream directly"

	^ self errorPipelineStream! !

!PipeableEvaluator methodsFor: 'printing' stamp: 'dtl 1/12/2003 00:45'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' ('; nextPutAll: self runState; nextPut: $)
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:37'!
runState

	^ runState ifNil: [runState := #notYetRunning]
! !

!PipeableEvaluator methodsFor: 'accessing' stamp: 'dtl 11/8/2007 20:37'!
runState: aSymbol

	runState := aSymbol.
	self changed: #runState
! !

!PipeableEvaluator methodsFor: 'setting run state' stamp: 'dtl 3/4/2002 22:32'!
running
	"Process is actively running."

	self runState: #running
! !

!PipeableEvaluator methodsFor: 'error handling' stamp: 'dtl 4/7/2001 21:27'!
selectFrom: start to: end

	^ self blockString copyFrom: start to: end! !

!PipeableEvaluator methodsFor: 'error handling' stamp: 'dtl 4/7/2001 19:50'!
selectionInterval
	"The compiler wants this for some reason."

	^ 1 to: 1! !

!PipeableEvaluator methodsFor: 'testing' stamp: 'dtl 12/22/2001 18:37'!
succeeded
	"Answer true if my evaluation completed successfully."

	^ self exitStatus == #success! !

!PipeableEvaluator methodsFor: 'evaluating' stamp: 'dtl 11/25/2001 22:39'!
success

	self exitStatus: #success! !

!PipeableEvaluator methodsFor: 'error handling' stamp: 'dtl 12/8/2001 18:13'!
text
	"The compiler needs this to generate error messages."

	^ Text fromString: self expression! !

!PipeableEvaluator methodsFor: 'setting run state' stamp: 'dtl 3/4/2002 22:32'!
unknownRunState
	"Unable to determine the current run state of the process."

	self runState: #unknownRunState
! !

!PipeableEvaluator methodsFor: 'command pipelines' stamp: 'dtl 11/8/2007 20:38'!
| anotherCommandString
	"Pipe my output to another evaluator, where the other evaluator evaluates
	anotherCommandString."

	"((ExpressionEvaluator command: 'Smalltalk')
		| 'stdout nextPutAll: stdin upToEnd'
		| 'stdout nextPutAll: stdin upToEnd') upToEnd"

	"(ExpressionEvaluator command: 'BOGUS')
		| 'stdout nextPutAll: stdin upToEnd'
		| 'stdout nextPutAll: stdin upToEnd'"

	"((ExpressionEvaluator command: '(true == false) not')
		| 'stdout nextPutAll: stdin upToEnd'
		| 'stdout nextPutAll: stdin upToEnd') upToEnd"

	| new |
	new := self species
			command: anotherCommandString
			pipeFrom: self pipeFromOutput
			errorPipelineStream: self errorPipelineStream.
	self value.
	^ new
! !

PipeableEvaluator subclass: #ShellBuiltinEvaluator
	instanceVariableNames: 'shellBuiltin commandShell'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CommandShell-Piping'!

!ShellBuiltinEvaluator commentStamp: 'dtl 2/1/2003 17:47' prior: 0!
I provide the context in which a ShellBuiltin is evaluated as an element of a command pipeline.!

!ShellBuiltinEvaluator class methodsFor: 'instance creation' stamp: 'dtl 2/1/2003 16:01'!
command: aShellBuiltin withArguments: argumentArray shell: commandShell pipeFrom: inputStream pipeTo: outputStream errorPipelineStream: errorStream

	^ super new
		shellBuiltin: aShellBuiltin;
		argumentArray: argumentArray;
		commandShell: commandShell;
		pipeToInput: inputStream;
		pipeFromOutput: outputStream;
		errorPipelineStream: errorStream;
		initialize! !

!ShellBuiltinEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 15:13'!
commandShell

	^ commandShell
! !

!ShellBuiltinEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 15:12'!
commandShell: aCommandShell

	commandShell := aCommandShell
! !

!ShellBuiltinEvaluator methodsFor: 'testing' stamp: 'dtl 2/1/2003 16:08'!
needsPrompt
	"Answer true if a shell will want to issue a prompt after evaluating me."

	^ self shellBuiltin needsPrompt
! !

!ShellBuiltinEvaluator methodsFor: 'printing' stamp: 'dtl 3/22/2003 13:58'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' on '.
	self shellBuiltin printOn: aStream! !

!ShellBuiltinEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 15:12'!
shellBuiltin

	^ shellBuiltin! !

!ShellBuiltinEvaluator methodsFor: 'accessing' stamp: 'dtl 2/1/2003 15:12'!
shellBuiltin: aShellBuiltin

	shellBuiltin := aShellBuiltin! !

!ShellBuiltinEvaluator methodsFor: 'evaluating' stamp: 'dtl 12/21/2007 12:50'!
value

	self isNotYetRunning ifTrue:
		[self success. "Hope for the best, send #fail if something goes wrong"
		self running.
		self ensurePipeFromOutput.
		self shellBuiltin
				performWith: self argumentArray
				evaluator: self
				shell: self commandShell
				input: self pipeToInput
				output: self pipeFromOutput
				error: self errorPipelineStream]
! !
